const axios = require('axios');

let instancesCreated = 0;
let instancesDestroyed = 0;

class Expression {
  constructor() {
    instancesCreated++;
  }

  toString() { return ""; }
  derive(variableName) { return new Nombre(0); }
  simplify() { return this; }
  valueOf() { return NaN; }
  destroy() { instancesDestroyed++; }
}

class Nombre extends Expression {
  constructor(value) {
    super();
    this.value = value;
  }

  toString() { return this.value.toString(); }
  derive(_) { return new Nombre(0); }
  simplify() { return this; }
  valueOf() { return this.value; }
}

class Variable extends Expression {
  constructor(name) {
    super();
    this.name = name;
  }

  toString() { return this.name; }
  derive(variableName) {
    return this.name === variableName ? new Nombre(1) : new Nombre(0);
  }
  simplify() { return this; }
  valueOf() { return NaN; }
}

class Operation extends Expression {
  constructor(leftOperand, rightOperand) {
    super();
    this.leftOperand = leftOperand;
    this.rightOperand = rightOperand;
  }
}

class Addition extends Operation {
  constructor(leftOperand, rightOperand) {
    super(leftOperand, rightOperand);
  }

  toString() {
    return `(${this.leftOperand.toString()} + ${this.rightOperand.toString()})`;
  }

  derive(variableName) {
    return new Addition(
      this.leftOperand.derive(variableName),
      this.rightOperand.derive(variableName)
    );
  }

  simplify() {
    const left = this.leftOperand.simplify();
    const right = this.rightOperand.simplify();

    if (!isNaN(left.valueOf()) && !isNaN(right.valueOf())) {
      return new Nombre(left.valueOf() + right.valueOf());
    }

    if (!isNaN(left.valueOf()) && left.valueOf() === 0) return right;
    if (!isNaN(right.valueOf()) && right.valueOf() === 0) return left;

    return new Addition(left, right);
  }

  valueOf() {
    const l = this.leftOperand.valueOf();
    const r = this.rightOperand.valueOf();
    return isNaN(l) || isNaN(r) ? NaN : l + r;
  }
}

class Multiplication extends Operation {
  constructor(leftOperand, rightOperand) {
    super(leftOperand, rightOperand);
  }

  toString() {
    return `(${this.leftOperand.toString()} * ${this.rightOperand.toString()})`;
  }

  derive(variableName) {
    return new Addition(
      new Multiplication(this.leftOperand.derive(variableName), this.rightOperand),
      new Multiplication(this.leftOperand, this.rightOperand.derive(variableName))
    );
  }

  simplify() {
    const left = this.leftOperand.simplify();
    const right = this.rightOperand.simplify();

    if (!isNaN(left.valueOf()) && !isNaN(right.valueOf())) {
      return new Nombre(left.valueOf() * right.valueOf());
    }

    if ((!isNaN(left.valueOf()) && left.valueOf() === 0) ||
        (!isNaN(right.valueOf()) && right.valueOf() === 0)) {
      return new Nombre(0);
    }

    if (!isNaN(left.valueOf()) && left.valueOf() === 1) return right;
    if (!isNaN(right.valueOf()) && right.valueOf() === 1) return left;

    return new Multiplication(left, right);
  }

  valueOf() {
    const l = this.leftOperand.valueOf();
    const r = this.rightOperand.valueOf();
    return isNaN(l) || isNaN(r) ? NaN : l * r;
  }
}

// ==== Construction asynchrone 

async function getContent() {
  const res = await axios.get('http://wdi.centralesupelec.fr/expression');
  console.log("Réponse reçue du serveur :", res.data);
  return res.data.content;
}

async function buildExpressionFromServer() {
  const content = await getContent();

  if (!isNaN(parseInt(content))) {
    return new Nombre(parseInt(content));
  }

  if (content === 'x' || content === 'y') {
    return new Variable(content);
  }

  if (content === '+' || content === '*') {
    const left = await buildExpressionFromServer();
    const right = await buildExpressionFromServer();
    return content === '+' ?
      new Addition(left, right) :
      new Multiplication(left, right);
  }

  throw new Error("Contenu inattendu : " + content);
}

// tick pour asynchronisme

let running = true;

async function tick() {
  const sleep = ms => new Promise(r => setTimeout(r, ms));
  while (running) {
    console.log('Sleeping 0.2 second...');
    await sleep(200);
  }
}

async function main() {
  tick(); // Lancer le suivi asynchrone

  const expr = await buildExpressionFromServer();
  console.log("Expression construite :", expr.toString());

  const dx = expr.derive("x");
  const dy = expr.derive("y");

  console.log("Dérivée par rapport à x :", dx.toString());
  console.log("Dérivée simplifiée (x) :", dx.simplify().toString());

  console.log("Dérivée par rapport à y :", dy.toString());
  console.log("Dérivée simplifiée (y) :", dy.simplify().toString());

  running = false; // Arrêt du tick
}

main();


// Sleeping 0.2 second...
// Sleeping 0.2 second...
// Sleeping 0.2 second...
// Réponse reçue du serveur : { content: '*' }
// Sleeping 0.2 second...
// Sleeping 0.2 second...
// Réponse reçue du serveur : { content: '*' }
// Sleeping 0.2 second...
// Sleeping 0.2 second...
// Sleeping 0.2 second...
// Réponse reçue du serveur : { content: '3' }
// Sleeping 0.2 second...
// Sleeping 0.2 second...
// Sleeping 0.2 second...
// Sleeping 0.2 second...
// Réponse reçue du serveur : { content: '+' }
// Sleeping 0.2 second...
// Sleeping 0.2 second...
// Sleeping 0.2 second...
// Réponse reçue du serveur : { content: '+' }
// Sleeping 0.2 second...
// Sleeping 0.2 second...
// Sleeping 0.2 second...
// Réponse reçue du serveur : { content: 'x' }
// Sleeping 0.2 second...
// Sleeping 0.2 second...
// Sleeping 0.2 second...
// Réponse reçue du serveur : { content: '*' }
// Sleeping 0.2 second...
// Sleeping 0.2 second...
// Sleeping 0.2 second...
// Sleeping 0.2 second...
// Sleeping 0.2 second...
// Réponse reçue du serveur : { content: 'y' }
// Sleeping 0.2 second...
// Sleeping 0.2 second...
// Réponse reçue du serveur : { content: '6' }
// Sleeping 0.2 second...
// Sleeping 0.2 second...
// Sleeping 0.2 second...
// Réponse reçue du serveur : { content: '0' }
// Sleeping 0.2 second...
// Sleeping 0.2 second...
// Sleeping 0.2 second...
// Sleeping 0.2 second...
// Réponse reçue du serveur : { content: 'x' }
// Expression construite : ((3 * ((x + (y * 6)) + 0)) * x)
// Dérivée par rapport à x : ((((0 * ((x + (y * 6)) + 0)) + (3 * ((1 + ((0 * 6) + (y * 0))) + 0))) * x) + ((3 * ((x + (y * 6)) + 0)) * 1))
// Dérivée simplifiée (x) : ((3 * x) + (3 * (x + (y * 6))))
// Dérivée par rapport à y : ((((0 * ((x + (y * 6)) + 0)) + (3 * ((0 + ((1 * 6) + (y * 0))) + 0))) * x) + ((3 * ((x + (y * 6)) + 0)) * 0))
// Dérivée simplifiée (y) : (18 * x)