let instancesCreated = 0;
let instancesDestroyed = 0;

class Expression {
  constructor() {
    instancesCreated++;
  }
  
  toString() {
    return "";
  }
  

  derive(variableName) {
    return new Nombre(0);
  }
  
  simplify() {
    return this;
  }
  
  valueOf() {
    return NaN; 
  }
  

  destroy() {
    instancesDestroyed++;
  }
}

class Nombre extends Expression {
  constructor(value) {
    super();
    this.value = value;
  }
  
  toString() {
    return this.value.toString();
  }
  
  // dérivée d'un nombre constant est toujours 0
  derive(variableName) {
    return new Nombre(0);
  }
  
  simplify() {
    return this;
  }
  
  valueOf() {
    return this.value;
  }
}

class Variable extends Expression {
  constructor(name) {
    super();
    this.name = name;
  }
  
  toString() {
    return this.name;
  }
  
  // la dérivée d'une variable est 1 si c'est la variable par rapport à laquelle
  // on dérive, 0 sinon
  derive(variableName) {
    return this.name === variableName ? new Nombre(1) : new Nombre(0);
  }
  
  simplify() {
    return this;
  }
  
  valueOf() {
    return NaN;
  }
}

class Operation extends Expression {
  constructor(leftOperand, rightOperand) {
    super();
    this.leftOperand = leftOperand;
    this.rightOperand = rightOperand;
  }
}

class Addition extends Operation {
  constructor(leftOperand, rightOperand) {
    super(leftOperand, rightOperand);
  }
  
  toString() {
    return `(${this.leftOperand.toString()} + ${this.rightOperand.toString()})`;
  }
  
  // dérivée d'une somme = somme des dérivées
  derive(variableName) {
    return new Addition(
      this.leftOperand.derive(variableName),
      this.rightOperand.derive(variableName)
    );
  }
  
  simplify() {
    const leftSimplified = this.leftOperand.simplify();
    const rightSimplified = this.rightOperand.simplify();
    
    // si les deux opérandes peuvent être évalués numériquement, on calcule le résultat
    if (!isNaN(leftSimplified.valueOf()) && !isNaN(rightSimplified.valueOf())) {
      return new Nombre(leftSimplified.valueOf() + rightSimplified.valueOf());
    }
    
    // si l'un des opérandes est 0, on retourne l'autre
    if (!isNaN(leftSimplified.valueOf()) && leftSimplified.valueOf() === 0) {
      return rightSimplified;
    }
    if (!isNaN(rightSimplified.valueOf()) && rightSimplified.valueOf() === 0) {
      return leftSimplified;
    }
    
    // sinon, on crée une nouvelle addition avec les opérandes simplifiés
    return new Addition(leftSimplified, rightSimplified);
  }
  
  // une addition peut être évaluée si ses deux opérandes peuvent l'être
  valueOf() {
    const leftValue = this.leftOperand.valueOf();
    const rightValue = this.rightOperand.valueOf();
    
    if (isNaN(leftValue) || isNaN(rightValue)) {
      return NaN;
    }
    return leftValue + rightValue;
  }
}

class Multiplication extends Operation {
  constructor(leftOperand, rightOperand) {
    super(leftOperand, rightOperand);
  }
  
  toString() {
    return `(${this.leftOperand.toString()} * ${this.rightOperand.toString()})`;
  }
  
  // (f*g)' = f'*g + f*g'
  derive(variableName) {
    const leftDerivative = this.leftOperand.derive(variableName);
    const rightDerivative = this.rightOperand.derive(variableName);
    
    return new Addition(
      new Multiplication(leftDerivative, this.rightOperand),
      new Multiplication(this.leftOperand, rightDerivative)
    );
  }
  
  // simplification d'une multiplication
  simplify() {
    const leftSimplified = this.leftOperand.simplify();
    const rightSimplified = this.rightOperand.simplify();
    
    // si les deux opérandes peuvent être évalués numériquement, on calcule le résultat
    if (!isNaN(leftSimplified.valueOf()) && !isNaN(rightSimplified.valueOf())) {
      return new Nombre(leftSimplified.valueOf() * rightSimplified.valueOf());
    }
    
    // si l'un des opérandes est 0, le résultat est 0
    if ((!isNaN(leftSimplified.valueOf()) && leftSimplified.valueOf() === 0) ||
        (!isNaN(rightSimplified.valueOf()) && rightSimplified.valueOf() === 0)) {
      return new Nombre(0);
    }
    
    // si l'un des opérandes est 1, on retourne l'autre
    if (!isNaN(leftSimplified.valueOf()) && leftSimplified.valueOf() === 1) {
      return rightSimplified;
    }
    if (!isNaN(rightSimplified.valueOf()) && rightSimplified.valueOf() === 1) {
      return leftSimplified;
    }
    
    // sinon, on crée une nouvelle multiplication avec les opérandes simplifiés
    return new Multiplication(leftSimplified, rightSimplified);
  }
  
  // multiplication peut être évaluée si ses deux opérandes peuvent l'être
  valueOf() {
    const leftValue = this.leftOperand.valueOf();
    const rightValue = this.rightOperand.valueOf();
    
    if (isNaN(leftValue) || isNaN(rightValue)) {
      return NaN;
    }
    
    return leftValue * rightValue;
  }
}



function test() {
  const x = new Variable("x");
  const y = new Variable("y");
  const deux = new Nombre(2);
  const trois = new Nombre(3);
  
  // expression: 2*x + 3*y
  const expr = new Addition(
    new Multiplication(deux, x),
    new Multiplication(trois, y)
  );
  
  console.log("Expression originale:", expr.toString());
  
  // dérivation par rapport à x
  const derivedX = expr.derive("x");
  console.log("Dérivée par rapport à x:", derivedX.toString());
  
  // dérivation par rapport à y
  const derivedY = expr.derive("y");
  console.log("Dérivée par rapport à y:", derivedY.toString());
  
  // simplification de la dérivée par rapport à x
  const simplifiedDerivedX = derivedX.simplify();
  console.log("Dérivée simplifiée par rapport à x:", simplifiedDerivedX.toString());
  
  // simplification de la dérivée par rapport à y
  const simplifiedDerivedY = derivedY.simplify();
  console.log("Dérivée simplifiée par rapport à y:", simplifiedDerivedY.toString());
  
  
  // Tests de valueOf()
  console.log("\nTests de valueOf():");
  console.log("Valeur de 2:", deux.valueOf());
  console.log("Valeur de x:", x.valueOf());
  console.log("Valeur de 2+3:", new Addition(deux, trois).valueOf());
  console.log("Valeur de 2*3:", new Multiplication(deux, trois).valueOf());
  console.log("Valeur de x+y:", new Addition(x, y).valueOf());
}

// Exécution du test
test();